%% Dataset.
% Training data are named X1 and X2 for view 1/2.
% Tuning data are named XV1 and XV2 for view 1/2.
% Testing data are named XTe1 and XTe2 for view 1/2.
% All data matrices contains samples rowwise.
% 
% The dataset of halved MNIST images used in Andrew et al. 2013 can be 
%   generated using the script createMNIST.m and the mat file mnist_all.mat
%   downloaded from Sam Roweis's webpage
%   http://www.cs.nyu.edu/~roweis/data.html
% 
% Uncomment the following line if you want to generate the dataset.
% createMNIST;

clear
load MNIST.mat

%% Final CCA dimension.
K=50;
%% Regularizations for each view.
rcov1=1e-4; rcov2=1e-4;

%% Run linear CCA.
[P1,P2,m1,m2,~]=linCCA(X1,X2,K,rcov1,rcov2);
% Testing the linear CCA projections.
X1proj=bsxfun(@minus,X1,m1)*P1;
X2proj=bsxfun(@minus,X2,m2)*P2;
fprintf('Linear CCA: training correlation %f\n',DCCA_corr(X1proj,X2proj,K));
XV1proj=bsxfun(@minus,XV1,m1)*P1;
XV2proj=bsxfun(@minus,XV2,m2)*P2;
fprintf('Linear CCA: tuning correlation %f\n',DCCA_corr(XV1proj,XV2proj,K));
XTe1proj=bsxfun(@minus,XTe1,m1)*P1;
XTe2proj=bsxfun(@minus,XTe2,m2)*P2;
fprintf('Linear CCA: testing correlation %f\n\n',DCCA_corr(XTe1proj,XTe2proj,K));

%% Run randomized kernel CCA.
% Number of random Fourier features to use.
M1=6000;  M2=6000;
% Kernel widths.
s1=16;  s2=16;
[W1,b1,W2,b2,P1,P2,m1,m2,~,randseed]=randKCCA(X1,X2,K,M1,s1,rcov1,M2,s2,rcov2);
% Testing the learned RKCCA projections.
X1proj=bsxfun(@minus,cos(bsxfun(@plus,X1*W1,b1)),m1)*P1;
X2proj=bsxfun(@minus,cos(bsxfun(@plus,X2*W2,b2)),m2)*P2;
fprintf('Randomized KCCA: training correlation %f\n',DCCA_corr(X1proj,X2proj,K));
XV1proj=bsxfun(@minus,cos(bsxfun(@plus,XV1*W1,b1)),m1)*P1;
XV2proj=bsxfun(@minus,cos(bsxfun(@plus,XV2*W2,b2)),m2)*P2;
% You should get a total canonical correlation of 43.1 on tuning set.
fprintf('Randomized KCCA: tuning correlation %f\n',DCCA_corr(XV1proj,XV2proj,K));
XTe1proj=bsxfun(@minus,cos(bsxfun(@plus,XTe1*W1,b1)),m1)*P1;
XTe2proj=bsxfun(@minus,cos(bsxfun(@plus,XTe2*W2,b2)),m2)*P2;
% You should get a total canonical correlation of 43.1 on test set.
fprintf('Randomized KCCA: testing correlation %f\n\n',DCCA_corr(XTe1proj,XTe2proj,K));

%% Hyperparameters for DCCA network architecture.
% Hidden activation type.
hiddentype='relu';
% Architecture (hidden layer sizes) for view 1 neural network.
NN1=[800 800 K];
% Architecture (hidden layer sizes)  for view 2 neural network.
NN2=[800 800 K];
% Weight decay parameter.
l2penalty=1e-3;

%% Run DCCA with SGD.
% Minibatchsize.
batchsize=500;
% Learning rate.
eta0=0.001;
% Rate at which learning rate decays over iterations.
% 1 means constant learning rate.
decay=1;
% Momentum.
momentum=0.99;
% How many passes over the data to run SGD with.
maxepoch=50;
addpath ./deepnet/
[F1opt,F2opt]=DCCAtrain_SGD(X1,X2,XV1,XV2,[],[],K,hiddentype,NN1,NN2, ...
  rcov1,rcov2,l2penalty,batchsize,eta0,decay,momentum,maxepoch);
% Testing the learned networks.
X1proj=deepnetfwd(X1,F1opt);  X2proj=deepnetfwd(X2,F2opt);
fprintf('Deep CCA (SGD): training correlation %f\n',DCCA_corr(X1proj,X2proj,K));
XV1proj=deepnetfwd(XV1,F1opt);  XV2proj=deepnetfwd(XV2,F2opt);
% You should get a total canonical correlation of 47.2 on tuning set.
fprintf('Deep CCA (SGD): tuning correlation %f\n',DCCA_corr(XV1proj,XV2proj,K));
XTe1proj=deepnetfwd(XTe1,F1opt);  XTe2proj=deepnetfwd(XTe2,F2opt);
% You should get a total canonical correlation of 47.2 on test set.
fprintf('Deep CCA (SGD): testing correlation %f\n\n',DCCA_corr(XTe1proj,XTe2proj,K));

%% Run DCCA with LBFGS.
addpath ./deepnet/;
addpath ./deepnet/LBFGS/minFunc/;
addpath ./deepnet/LBFGS/minFunc/compiled/;
% Maximum number of iterations.
maxit=300;
% Number of past steps for estimating Hessian in L-BFGS.
paststeps=5;
[F1opt,F2opt]=DCCAtrain_LBFGS(X1,X2,XV1,XV2,[],[],K,hiddentype,NN1,NN2, ...
  rcov1,rcov2,l2penalty,maxit,paststeps);
% Testing the learned networks.
X1proj=deepnetfwd(X1,F1opt);  X2proj=deepnetfwd(X2,F2opt);
fprintf('Deep CCA (L-BFGS): training correlation %f\n',DCCA_corr(X1proj,X2proj,K));
XV1proj=deepnetfwd(XV1,F1opt);  XV2proj=deepnetfwd(XV2,F2opt);
% You should get a total canonical correlation of 47.3 on tuning set.
fprintf('Deep CCA (L-BFGS): tuning correlation %f\n',DCCA_corr(XV1proj,XV2proj,K));
XTe1proj=deepnetfwd(XTe1,F1opt);  XTe2proj=deepnetfwd(XTe2,F2opt);
% You should get a total canonical correlation of 47.2 on test set.
fprintf('Deep CCA (L-BFGS): testing correlation %f\n\n',DCCA_corr(XTe1proj,XTe2proj,K));
